// Copyright (c) <2012> <Leif Asbrink>
//
// Permission is hereby granted, free of charge, to any person 
// obtaining a copy of this software and associated documentation 
// files (the "Software"), to deal in the Software without restriction, 
// including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of 
// the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be 
// included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
// WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE 
// OR OTHER DEALINGS IN THE SOFTWARE.


#if LIBRTLSDR_PRESENT == 1
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "globdef.h"
#include "uidef.h"
#include "sdrdef.h"
#include "conf.h"
#include "fft1def.h"
#include "screendef.h"
#include "rusage.h"
#include "thrdef.h"
#include "options.h"
#include "vernr.h"
#include "hwaredef.h"

#if(OSNUM == OSNUM_WINDOWS)
#include <windows.h>
#endif

#if(OSNUM == OSNUM_LINUX)
#include <pthread.h>
#include <unistd.h>
#endif

#include <rtl-sdr.h>

#define MAX_RTL2832_SAMP_RATE_H  2400000
#define MIN_RTL2832_SAMP_RATE_H  900001
#define MAX_RTL2832_SAMP_RATE_L  300000
#define MIN_RTL2832_SAMP_RATE_L  230000

P_RTL2832 rtl2832;

static rtlsdr_dev_t *dev_rtlsdr;

char *rtl2832_parm_text[MAX_RTL2832_PARM]={"Sampling speed",
                                     "Serno1",
                                     "Serno2",
                                     "Serno3",
                                     "Gain mode",
                                     "Freq adjust",
                                     "Direct sampling",
                                     "Check"};

char *rtl2832_parfil_name="par_rtl2832";

static  void rtlsdr_callback(unsigned char *buf, uint32_t len, void *ctx)
{
(void) ctx;
unsigned int i;
short int *iz;
iz=(short int*)&timf1_char[timf1p_sdr];
for(i=0; i<len; i++)
  {
  iz[i]=(buf[i]+127)<<8;
  }
timf1p_sdr=(timf1p_sdr+(len<<1))&timf1_bytemask;
if( ((timf1p_sdr-timf1p_pa+timf1_bytes)&timf1_bytemask) >= 
                                                     snd[RXAD].block_bytes)
  {
  lir_set_event(EVENT_HWARE1_RXREADY);
  }
}

void set_rtl2832_att(void)
{
if(rtl2832.gain_mode == 0)return;
rtlsdr_set_tuner_gain(dev_rtlsdr, old_rtl2832_gain);
}

void set_rtl2832_frequency(void)
{
uint32_t frequency;
frequency=(fg.passband_center*(100000000-rtl2832.freq_adjust))/100;
rtlsdr_set_center_freq(dev_rtlsdr, frequency);
}

void rtl_starter(void)
{
int k, no_of_buffers;
k=snd[RXAD].block_bytes/2;
if(k < 512)k=512;
// Allocate buffers for 50 ms minimum.
no_of_buffers=(int)((0.1*rtl2832.sampling_speed)/k);
// Never use less than 4 buffers.
if(no_of_buffers < 4)no_of_buffers=4;
while(thread_status_flag[THREAD_RTL2832_INPUT]!=THRFLAG_ACTIVE)
  {
  lir_sleep(10000);
  if(kill_all_flag)return;
  }
rtlsdr_read_async(dev_rtlsdr, rtlsdr_callback, NULL,
				      no_of_buffers, k);
}

void rtl2832_input(void)
{
#if RUSAGE_OLD == TRUE
int local_workload_counter;
#endif
int i, j;
int rxin_local_workload_reset;
char s[80];
double dt1, read_start_time, total_reads;
int timing_loop_counter,timing_loop_counter_max,initial_skip_flag;
int local_att_counter;
int local_nco_counter;
char vendor[256], product[256], serial[256];
int *ise;
unsigned int no_of_rtl;
uint32_t idx;
lir_init_event(EVENT_HWARE1_RXREADY);
#if OSNUM == OSNUM_LINUX
clear_thread_times(THREAD_RTL2832_INPUT);
#endif
#if RUSAGE_OLD == TRUE
local_workload_counter=workload_counter;
#endif
local_att_counter=sdr_att_counter;
local_nco_counter=sdr_nco_counter;
dt1=current_time();
i=read_sdrpar(rtl2832_parfil_name, MAX_RTL2832_PARM, 
                                     rtl2832_parm_text, (int*)((void*)&rtl2832));
if(i != 0 || rtl2832.check != RTL2832PAR_VERNR)
  {
  lirerr(1356);
  goto rtl2832_init_error_exit;
  } 
if(ui.rx_ad_speed != (int)((rint)((rtl2832.sampling_speed*
                     (100000000L+(double)rtl2832.freq_adjust))/100000000L)))
  {
  lirerr(1356);
  goto rtl2832_init_error_exit;
  } 
timf1p_sdr=timf1p_pa;  
j=0;
while(sdr == -1)
  {
  ise=(int*)(void*)s;
  dev_rtlsdr=NULL;
  no_of_rtl = rtlsdr_get_device_count();
  for(idx=0; idx<no_of_rtl; idx++)
    {
    rtlsdr_get_device_usb_strings(idx, vendor, product, serial);
    sprintf(s,"%s",serial);
    i=0;     
    while(i<16 && s[i]!=0)i++;
    while(i<16 )
      {
      s[i]=' ';
      i++;
      }
    if( rtl2832.sernum1 == ise[0] &&
        rtl2832.sernum2 == ise[1] &&
        rtl2832.sernum3 == ise[2])
      {  
      sdr=rtlsdr_open(&dev_rtlsdr, idx);
      if(kill_all_flag)  goto rtl2832_init_error_exit;
      if(sdr >= 0)
        {
        i=rtlsdr_set_sample_rate(dev_rtlsdr, (uint32_t)rtl2832.sampling_speed);
        if(i < 0)
          {
          lirerr(1339);
opnerr:;        
          rtlsdr_close(dev_rtlsdr);
          goto rtl2832_init_error_exit;
          }
// Direct Sampling mode. tuner = 0 (default), I = 1, Q = 2
        if(rtl2832.direct_sampling>0)
          {
          i=rtlsdr_set_direct_sampling(dev_rtlsdr, rtl2832.direct_sampling);
          if(i!=0)
            {
            lirerr(1358);
            goto opnerr; 
            }
// Enable rtl2831 AGC
          if(rtl2832.gain_mode==0) i=rtlsdr_set_agc_mode(dev_rtlsdr, 1);
          }
// Reset endpoint before we start reading from the RTL2832 (mandatory)
        i=rtlsdr_reset_buffer(dev_rtlsdr);
        if(i < 0)
          {
          lirerr(1340);
          goto opnerr;
          }
        linrad_thread_create(THREAD_RTL_STARTER);
        i=rtlsdr_set_tuner_gain_mode(dev_rtlsdr, rtl2832.gain_mode);
        if(i != 0 && i != rtl2832.gain_mode)goto rtl2832_error_exit;
        if(rtl2832.gain_mode != 0)
          {
          no_of_rtl2832_gains=rtlsdr_get_tuner_gains(dev_rtlsdr, NULL);
          if(no_of_rtl2832_gains < 1 ||
             no_of_rtl2832_gains > 256)goto rtl2832_error_exit;
          rtl2832_gains=malloc(no_of_rtl2832_gains*sizeof(int));
          if(rtl2832_gains == NULL)goto rtl2832_error_exit;
          rtlsdr_get_tuner_gains(dev_rtlsdr, rtl2832_gains);
          }
        break;
        }
      else
        {
        if(j==0)
          {
          clear_screen();
          j=1;
          }
        }
      }
    }
  sprintf(s,"Waiting %.2f", current_time()-dt1);
  lir_text(3,5,s);
  lir_refresh_screen();
  if(kill_all_flag)goto rtl2832_init_error_exit;
  lir_sleep(100000);
  }
set_hardware_rx_gain();
set_rtl2832_att();
set_rtl2832_frequency();
fft1_block_timing();
fft1_hz_per_point=(float)ui.rx_ad_speed/fft1_size;
if(thread_command_flag[THREAD_SCREEN]!=THRFLAG_NOT_ACTIVE)
  {
  while(thread_status_flag[THREAD_SCREEN]!=THRFLAG_ACTIVE &&
        thread_status_flag[THREAD_SCREEN]!=THRFLAG_IDLE &&
        thread_status_flag[THREAD_SCREEN]!=THRFLAG_SEM_WAIT)
    {
    if(thread_command_flag[THREAD_RTL2832_INPUT] ==
                                           THRFLAG_KILL)goto rtl2832_error_exit1;
    lir_sleep(10000);
    }
  }
#include "timing_setup.c"
thread_status_flag[THREAD_RTL2832_INPUT]=THRFLAG_ACTIVE;
while(!kill_all_flag && 
            thread_command_flag[THREAD_RTL2832_INPUT] == THRFLAG_ACTIVE)
  {
#if RUSAGE_OLD == TRUE
  if(local_workload_counter != workload_counter)
    {
    local_workload_counter=workload_counter;
    make_thread_times(THREAD_RTL2832_INPUT);
    }
#endif
  if(local_att_counter != sdr_att_counter)
    {
    local_att_counter=sdr_att_counter;
    set_rtl2832_att();
    }
  if(local_nco_counter != sdr_nco_counter)
    {
    local_nco_counter=sdr_nco_counter;
    set_rtl2832_frequency();
    }
  lir_await_event(EVENT_HWARE1_RXREADY);
  if(kill_all_flag)goto rtl2832_error_exit1;
  
  while (!kill_all_flag && timf1p_sdr != timf1p_pa)
    {
#include "input_speed.c"
    finish_rx_read();
    }
  }
rtl2832_error_exit1:;
if(rtl2832.gain_mode != 0)free(rtl2832_gains);
rtl2832_error_exit:;
rtlsdr_cancel_async(dev_rtlsdr);
rtlsdr_close(dev_rtlsdr);
lir_join(THREAD_RTL_STARTER);
rtl2832_init_error_exit:;
sdr=-1;
thread_status_flag[THREAD_RTL2832_INPUT]=THRFLAG_RETURNED;
while(thread_command_flag[THREAD_RTL2832_INPUT] != THRFLAG_NOT_ACTIVE)
  {
  lir_sleep(1000);
  }
lir_close_event(EVENT_HWARE1_RXREADY);
} 

void init_rtl2832(void)
{
FILE *rtl2832_file;
char s[120];
char vendor[256], product[256], serial[256];
int *ise, *sdr_pi;
int i, line, no_of_rtl, devno;
ise=(int*)(void*)s;
lir_text(3,2,"SEARCHING");
no_of_rtl = rtlsdr_get_device_count();
clear_lines(2,2);
if (no_of_rtl == 0) 
  {
  lir_text(5,5,"RTL2832 not found.");
  lir_text(5,7,press_any_key);
  await_keyboard();
  return;
  }
line=2;  
for(i=0; i<no_of_rtl; i++)
  {
  rtlsdr_get_device_usb_strings(i, vendor, product, serial);
  sprintf(s," %2d: %s   MFG:%s, PROD:%s, SN: %s", i, 
                           rtlsdr_get_device_name(i), vendor, product, serial);
  lir_text(3,line,s);
  line++;
  }
line++;
if (no_of_rtl == 1) 
  {
  lir_text(3,line,"Device autoselected.");
  devno=0;
  }
else
  {  
  lir_text(3, line, "Select device by line number:");
  devno=lir_get_integer(32,line,2,0,no_of_rtl-1);
  }
if(kill_all_flag)return;
rtlsdr_get_device_usb_strings(devno, vendor, product, serial);
sprintf(s,"%s",serial);
i=0;
while(i<16 && s[i]!=0)i++;
while(i<16)
  {
  s[i]=' ';
  i++;
  }
rtl2832.sernum1=ise[0];
rtl2832.sernum2=ise[1];
rtl2832.sernum3=ise[2];
line+=2;
lir_sleep(100000);
i=rtlsdr_open(&dev_rtlsdr, (uint32_t)devno);
if(i<0)
  {
#if(OSNUM == OSNUM_WINDOWS)
  lirerr(1348);
#else
  lirerr(1357);
#endif
  return;
  }
getspeed:;
sprintf(s,"Set sampling speed in Hz (%d to %d or %d to %d) =>",
                                    MIN_RTL2832_SAMP_RATE_L,MAX_RTL2832_SAMP_RATE_L,
                                    MIN_RTL2832_SAMP_RATE_H,MAX_RTL2832_SAMP_RATE_H);
lir_text(3,line,s);
rtl2832.sampling_speed=lir_get_integer(4+strlen(s),line,7,
                                MIN_RTL2832_SAMP_RATE_L,MAX_RTL2832_SAMP_RATE_H);
if(kill_all_flag)
  {
  rtlsdr_close(dev_rtlsdr);
  return;
  }
line+=2;
if(rtl2832.sampling_speed > MAX_RTL2832_SAMP_RATE_L &&
   rtl2832.sampling_speed < MIN_RTL2832_SAMP_RATE_H)
  {
  if( rtl2832.sampling_speed-MAX_RTL2832_SAMP_RATE_L >
      MIN_RTL2832_SAMP_RATE_H-rtl2832.sampling_speed)
    {
    rtl2832.sampling_speed=MIN_RTL2832_SAMP_RATE_H;
    }
  else       
    {
    rtl2832.sampling_speed=MAX_RTL2832_SAMP_RATE_L;
    }
  }  
i=rtlsdr_set_sample_rate(dev_rtlsdr, (uint32_t)rtl2832.sampling_speed);
rtl2832.sampling_speed=rtlsdr_get_sample_rate(dev_rtlsdr);
if(i < 0)
  {
  lir_text(3,line,"Failed to set sampling rate");
  lir_text(3,line+1,press_any_key);
  await_keyboard();
  clear_lines(line,line+1);
  goto getspeed;
  }
lir_text(3, line, "RF input modes:");
line++;
lir_text(3, line, "0 = Tuner");
line++;
lir_text(3, line, "1 = I Direct Sampling");
line++;
lir_text(3, line, "2 = Q Direct Sampling");
line+=2;
lir_text(3, line, "Select :");
rtl2832.direct_sampling=lir_get_integer(12,line,2,0,2);
if(kill_all_flag)
  {
  rtlsdr_close(dev_rtlsdr);
  return;
  }
if(rtl2832.direct_sampling==0)
  {
  i=rtlsdr_set_tuner_gain_mode(dev_rtlsdr, 32);
  } 
else 
  {
  i=1;
  }
if(i==0)i=1;
lir_text(3,line,"Select gain mode:");
line++;
lir_text(3,line,"0 = Auto");
line++;
lir_text(3,line,"1 = Original Osmocom");
line++;
if(i>1)
  {
  lir_text(3,line,"2 = Linearity");
  line++;
    {
    if(i > 3)
    lir_text(3,line,"to");
    line++;
    }
  sprintf(s,"%d = Sensitivity",i);
  lir_text(3,line,s);
  }
line++;
sprintf(s,"Enter gain mode (0 to %d) =>",i);
rtlsdr_close(dev_rtlsdr);
lir_text(3,line,s);
rtl2832.gain_mode=lir_get_integer(4+strlen(s),line,1,0,i);
if(kill_all_flag)return;
line+=2;
lir_text(3,line,"Enter xtal error in ppb =>");
rtl2832.freq_adjust=0.1*lir_get_float(32,line,9,-300000.,300000.);
if(kill_all_flag)return;
rtl2832_file=fopen(rtl2832_parfil_name,"w");
if(rtl2832_file == NULL)
  {
  lirerr(381268);
  return;
  }
rtl2832.check=RTL2832PAR_VERNR;
sdr_pi=(int*)(&rtl2832);
for(i=0; i<MAX_RTL2832_PARM; i++)
  {
  fprintf(rtl2832_file,"%s [%d]\n",rtl2832_parm_text[i],sdr_pi[i]);
  }
parfile_end(rtl2832_file);
ui.rx_addev_no=RTL2832_DEVICE_CODE;
ui.rx_ad_speed=(int)((rint)((rtl2832.sampling_speed*
                   (100000000L+(double)rtl2832.freq_adjust))/100000000L));
ui.rx_input_mode=IQ_DATA;
ui.rx_rf_channels=1;
ui.rx_ad_channels=2;
ui.rx_admode=0;
}

int display_rtl2832_parm_info(int *line)
{
char s[80];
int errcod;
char* input_modes[3]={"Tuner",
                      "I Direct Sampling",
                      "Q Direct Sampling"}; 
//if(FBDEV == 0)fflush(NULL);
errcod=read_sdrpar(rtl2832_parfil_name, MAX_RTL2832_PARM, 
                               rtl2832_parm_text, (int*)((void*)&rtl2832));
//if(FBDEV == 0)fflush(NULL);
if(errcod == 0)
  {
  settextcolor(7);
  sprintf(s,"Sampling rate = %i, Gain mode = %d, Xtal adjust = %.0f ppb",  
                              rtl2832.sampling_speed, rtl2832.gain_mode,
                              10.*rtl2832.freq_adjust);
  lir_text(24,line[0],s);
  SNDLOG"\n%s",s);
  line[0]++;
  sprintf(s,"Rf input mode = %d (%s)",rtl2832.direct_sampling,
                                 input_modes[rtl2832.direct_sampling]);
  lir_text(24,line[0],s);
  SNDLOG"\n%s",s);
  line[0]++;
  }
  
return (errcod);
}
#endif
